/*  srf_cutoff.c -- Determine cutoff distance for weighting function

NAME:
    srf_curtoff()

PURPOSE:
    Called by program SURFACE to iteratively determine the cutoff
    distance for the Gaussian weighting function at each point on a
    sparse grid covering the output image.

ARGUMENTS:
    long    *window	(in)	Input elevation image window.
    double  *obs	(in)	Array containing, for each observation,
				line and sample relative to  elevation
				image window, station elevation, and
				observation value.
    long    obscnt	(in)	Number of observation stations.
    long    niter	(in)	Number of iterations of cutoff
				calculation.
    long    stacnt	(in)	Target average sum of station weights.
    float   alpha	(in)	Shape parameter of Gaussian weighting
				function.
    FILE    *prtfp	(in)	File pointer for temporary report file.
    float   **cutptr	(out)	Pointer to array of cutoff distances at
				each sparse-grid point.
    long   *griddist	(out)	Grid spacing in sparse grid of cutoff
				distance.
    long    *ncol	(out)	Number of columns in sparse grid of
				cutoff distances.
    long    *nrow	(out)	Number of rows in sparse grid of
				cutoff distances.

RETURN VALUE:
    None.  Function aborts on error.

FILE REFERENCES:
    <*prtfp>		APPENDed to add to processing report.

PROGRAM HISTORY:                                  
    2001-07	Initial version, R. A. White, PSU/ESSC.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in surface.pdf.

ALGORITHM DESCRIPTION:
    Define box containing stations if any fall outside image window
    Determine greater of mean density of stations within and mean
      density within window
    Define sparse grid for defining cutoff distance with approximately
      25 cells per station
    For each point of sparse elevation grid
	Initalize trial cutoff distance
	For specified number of iterations
	    Determine weight for each station
	    Compute local effective station density
	    Update cutoff distance
	Reduce final cutoff distance by sqrt(0.5)
	Update extremat for cutoff, weight total, and station count
    Report grid spacing and dimensions
    Report minimum and maximum cutoff distance
    Report minimum and maximum sum of weights and station count

**************   End of Prolog for srf_cutoff()   ***************	*/
									/*
Header Files:			       					*/

#include	<math.h>	/* Standard math functions		*/
#include	<stdio.h>	/* Standard I/O functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
									/*
***************   Start of Function srf_cutoff   ***************	*/

void srf_cutoff(
    long    *window,	/* Input elevation image window */
    double  *obs,	/* Array containing, for each observation, line
			   and sample relative to  elevation image
			   window, station elevation, and observation
			   value */
    long    obscnt,	/* Number of observation stations */
    long    niter,	/* Number of iterations of cutoff calculation */
    long    stacnt,	/* Target average sum of station weights */
    float   alpha,	/* Shape parameter of Gaussian weighting
			   function */
    FILE    *prtfp,	/* File pointer for temporary report file */
    float   **cutptr,	/* Pointer to array of cutoff distances at each
			   sparse-grid point */
    long   *griddist,	/* Grid spacing in sparse grid of cutoff dist */
    long    *ncol,	/* Number of cols in sparse grid of cutoffs */ 
    long    *nrow)	/* Number of rows in sparse grid of cutoffs */
    {
				                  			/*
Local Variable Definitions:						*/

    float	avedens;	/* Average station density */
    float	b;		/* = sqrt( utweight * 2 * niter ) */
    long	col;		/* Current grid column */
    float	curcutoff;	/* Cutoff dist for current iteration */
    float	*cutoff;	/* Cutoff distance for each point on
				   sparse grid */
    float	cutmax;		/* Maximum cutoff distance */
    float	cutmin;		/* Minlimum cutoff distance */
    float	eminalpha;	/* Value of exp( - alpha ) */
    float	gausscoeff;	/* Coeff of dist squared in gaussian */ 
    long	i;		/* Loop counter/index */
    float	icutoff;	/* Initial value of cutoff distance */
    long	iter;		/* Current iteration counter */
    double	minmaxyx[4];	/* Boundaries of box for stations */
    long	nwithin;	/* Number of stations within window */
    float	pi;		/* Value of pi */
    long	row;		/* Current grid row */
    float	sqrdist;	/* Square of distance from station */
    float	sqrthalf;	/* Square root of 0.5 */
    long	staused = 0;	/* Number of stations used for point */
    long	staumax;	/* Maximum number of statons used */
    long	staumin;	/* Minimum number of statons used */
    float	totwmax;	/* Maxiumum total weights for point */
    float	totwmin;	/* Miniumum total weights for point */
    float	utweight;	/* Untruncated average weight */
    float	weightacc = 0.0;/* Accumulator for weights */
    float	x;		/* Sample number of current grid col. */ 
    float	y;		/* Line nunber of current grid row */

/*  Determine mean density of stations  */

    minmaxyx[0] = minmaxyx[2] = 1;
    minmaxyx[1] = window[2];
    minmaxyx[3] = window[3];
    for (i = 0, nwithin = 0; i < 4 * obscnt; i += 4)
	{
	if (obs[i] < minmaxyx[0])
	    minmaxyx[0] = obs[i];
	if (obs[i] > minmaxyx[1])
	    minmaxyx[1] = obs[i];
	if (obs[i + 1] < minmaxyx[2])
	    minmaxyx[2] = obs[i + 1];
	if (obs[i + 1] > minmaxyx[3])
	    minmaxyx[3] = obs[i + 1];
	if ( (obs[i] >= 1) && (obs[i] <= window[2]) && (obs[i + 1] >= 1)
	  && (obs[i + 1] <= window[3]) )
	    nwithin ++;
	}
    avedens = obscnt / ((minmaxyx[1] - minmaxyx[0] + 1)
      * (minmaxyx[3] - minmaxyx[2] + 1));
    if ((nwithin / (window[2] * window[3])) > avedens)
	avedens = nwithin / (window[2] * window[3]);

/*  Define sparse grid for refining cutoff distance  */

    *griddist = floor( sqrt( 1.0 / (25 * avedens) ) );
    if (*griddist < 1)
	*griddist = 1;
    *nrow = 2 + (window[2] - 1) / *griddist;
    *ncol = 2 + (window[3] - 1) / *griddist;
    if ((cutoff = (float *) malloc( *nrow * *ncol * sizeof( float ) ))
       == NULL)
	c_lasmsg( TRUE, "surface-malloc", "Memory allocation error");
    *cutptr = cutoff;

/*  Compute cutoff distance  */

    pi = 4 * atan2( 1.0, 1.0 );
    sqrthalf = sqrt( 0.5 );
    icutoff = sqrt( (2 * stacnt) / (avedens * pi) );
    eminalpha = exp( - alpha );
    utweight = ((1 - eminalpha) / alpha ) - eminalpha;
    b = sqrt( utweight * 2 * stacnt );
    cutmin = *griddist * (*ncol * *nrow);
    totwmin = staumin = obscnt;
    cutmax = totwmax = staumax = 0;
    for (row = 0, y = 1; row < *nrow; row++, y += *griddist)
	for (col = 0, x = 1; col < *ncol; col++, x += *griddist)
	    {
	    curcutoff = icutoff;
	    for (iter = 0; iter < niter; iter++)
		{
		weightacc = 0.0;
		staused = 0;
		gausscoeff = alpha / (curcutoff * curcutoff);
		for (i = 0; i < 4 * obscnt; i += 4)
		    {
		    sqrdist= (y - obs[i]) * (y - obs[i])
		      + (x - obs[i + 1]) * (x - obs[i + 1]);
		    if (sqrt( sqrdist ) <= curcutoff)
			{
			weightacc += exp( - gausscoeff * sqrdist ) -
			  eminalpha;
			staused++;
			}
		    }
		if (weightacc < 0.1)
		    weightacc = 0.1;
		curcutoff *= b * sqrt( 1 / weightacc );
		}
	    curcutoff *= sqrthalf;
	    cutoff[col + row * *ncol] = curcutoff;

/*  Check for extrema  */

	    if (curcutoff > cutmax)
		cutmax = curcutoff;
	    if (curcutoff < cutmin)
		cutmin = curcutoff;
	    if (weightacc > totwmax)
		totwmax = weightacc;
	    if (weightacc < totwmin)
		totwmin = weightacc;
	    if (staused > staumax)
		staumax = staused;
	    if (staused < staumin)
		staumin = staused;
	    }

/*  Report grid size and max/min weights and stations  */

    fprintf( prtfp, "\nCutoff distances defined for %ld %s %.2f\n",
      stacnt, "stations, alpha =", alpha );
    fprintf( prtfp, "    Coarse grid spacing %ld, %ld rows by %ld %s\n",
      *griddist, *nrow, *ncol, "columns" );
    fprintf( prtfp, "    Minimum and maximum cutoffs:  %.2f, %.2f\n",
      cutmin, cutmax );
    fprintf( prtfp, "    Minimum and maximum weights:  %.2f, %.2f\n",
      totwmin, totwmax );
    fprintf( prtfp, "    Minimum and maximum stations:  %ld, %ld\n",
      staumin, staumax );

    return;
    }			/* End of Function srf_cutoff() */
