/*  srf_getarg.c -- Get arguments for program SURFACE

NAME:
    srf_getarg()

PURPOSE:
    Called by SURFACE to get program arguments from the TAE executive.

ARGUMENTS:
    The arguments for SURFACE are defined in the prolog for surface.c.

    Arguments for srf_getarg are:a

    struct PARBLK
	    *vblock    (in/out)	Structure to receive arguments from TAE.
    char    *inelev	(out)	Hostname of input elevation image.
    long    window[4]	(out)	Input elevation image window.
    long    *band	(out)	Input elevation image band.
    char    *inobs	(out)	Name of station location and data file.
    char    *out	(out)	Hostname of output image.
    long    *odcode	(out)	Data type code for output image.
    float   *slope	(out)	Slope of regression line between
				variable and elevation.
    long    *slopeflag	(out)	Flag set TRUE if SLOPE specified.
    float   *alpha	(out)	Shape parameter of Gaussian weighting
				function.
    float   *stacnt	(out)	Target average number of stations for
				each grid point.
    long    *niter	(out)	Number of iterations for refining
				maximum distance.
    long    *regress	(out)	Code for regression type.
    float   *maxnd	(out)	Maximum absolute value of normalized
				difference.
    float   *minfract	(out)	Minimum fraction of stations above
				threshold for non-fill output.
    float   *thresh	(out)	Threshold for using variable value from
				station.
    float   *fill	(out)	Fill value for output pixels failing
				MINFRACT test.
    char    *prtfile	(out)	Name of temporary print file, if any.
    char    *listfile	(out)	Name of output listing file, if any.
    long    *pmsgtim	(out)	Time between processing progress messages.
    long    *ddrvflg	(out)   DDR validity option.

RETURN VALUE:
    None.  Function aborts if fatal error is encountered.

PROGRAM HISTORY:                                  
    2001-06	Initial version, R. A. White, PSU/ESSC.
    2005-07	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in surface.pdf.

ALGORITHM DESCRIPTION:

    Get program specific arguments
    Get TAE global variables
    Set up temporary file for optional header listing
    									
***************   End of Prolog for PROGRAM-NAME   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O functions		*/
#include	<string.h>	/* String processing functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"alias.h"	/* LAS evnironment variable function	*/
#include	"util.h"	/* LAS utility function definitions	*/
									/*
***************   Start of Function srf_getarg.c   ***************	*/

void srf_getarg(
    struct PARBLK
	    *vblock,	/* Structure to receive arguments from TAE */
    char    *inelev,	/* Hostname of input elevation image */
    long    window[4],  /* Input elevation image window */
    long    *band,	/* Input elevation image band */
    char    *inobs,	/* Name of station location and data file */
    char    *out,	/* Hostname of output image */
    long    *odcode,	/* Data type code for output image */
    float   *alpha,	/* Shape parameter of Gaussian weighting
			   function */
    float   *stacnt,	/* Target average number of stations for each
			   grid point */
    long    *niter,	/* Number of iterations for refining maximum
			   distance */
    float   *slope,	/* Slope of regression line between variable and
			   elevation */
    long    *slopeflag,	/* Flag set true if SLOPE specified */
    long    *regress,	/* Code for regression type */
    float   *maxnd,	/* Maximum absolute value of normalized
			   difference */
    float   *minfract,	/* Minimum fraction of stations above threshold
			   for non-fill output */
    float   *thresh,	/* Threshold for using variable value from
			   station */
    float   *fill,	/* Fill value for output pixels failing MINFRACT
			   test */
    char    *prtfile,	/* Name of temporary print file, if any */
    char    *listfile,	/* Name of output listing file, if any */
    long    *pmsgtim,	/* Time between processing progress messages */
    long    *ddrvflg)	/* DDR validity option */
    {
	                                              			/*
Local Variable Definitions:						*/

    char	argbuf[CMLEN];	/* Buffer for string argument */
    long	bands[2];	/* Array of bands needed by c_getin() */
    long	cmlen=CMLEN;	/* Constant equal to CMLEN */
    long	dummy[2];	/* Dummy arguments for c_getin() */
    long	elong=ELONG;	/* Constant equal to ELONG */
    long	ereal=EREAL;	/* Constant equal to EREAL */
    long	ifpfile;	/* Flag set if output print file */
    long	nmrimg;		/* Number of input images */
    long	one=1;		/* Constant value one */
    long	totband;	/* Total input bands */
    char	tname[CMLEN];	/* TAE name of image (ignored) */
    long	zero=0;		/* Constant value zero */

/*  Get input file arguments  */

    if ( (c_getstr( vblock, "INOBS", &one, &cmlen, inobs ) <= 0)
      || (c_getalias( inobs, inobs ) != E_SUCC) )
	c_lasmsg( TRUE, "surface-getarg",
	  "Error getting program arguments" );
    if (c_getin (vblock, "INELEV", &zero, (void *) tname,
      (void *) inelev, &nmrimg, (void *) window, (void *) bands, dummy,
      &totband, &one, dummy, (void *) dummy ) != E_SUCC) 
        c_lasmsg( TRUE, "surface-getarg",
	  "Error getting program arguments" );
    if ((nmrimg != 1) || (totband != 1))
	c_lasmsg( TRUE, "surface-oneband",
	  "Must specify single input elevation band" );
    *band = bands[0];

/*  Get output image name and type  */

    if ( c_getout( vblock, "OUT", (void *) tname, (void *) out )
       != E_SUCC)
	c_lasmsg( TRUE, "surface-getarg",
	  "Error getting program arguments" );
    if (c_getstr( vblock, "ODTYPE", &one, &cmlen, argbuf ) <= 0)
	c_lasmsg( TRUE, "surface-getarg",
	  "Error getting program arguments" );
    if (strcmp( argbuf, "BYTE" ) == 0)
        *odcode = EBYTE;
    else if (strcmp( argbuf, "I*2" ) == 0)
        *odcode = EWORD;
    else if (strcmp( argbuf, "I*4" ) == 0)
        *odcode = ELONG;
    else if (strcmp( argbuf, "R*4" ) == 0)
        *odcode = EREAL;
    else
        *odcode = 0;            /* Default to same as elevation grid */

/*  Get remaining program-specific arguments  */

    if (c_getnum( vblock, "SLOPE", &one, &ereal, (void *) slope ) > 0)
	*slopeflag = TRUE;
    else
	*slopeflag = FALSE;
    if ( (c_getnum( vblock, "ALPHA", &one, &ereal, (void *) alpha )
        <= 0) 
      || (c_getnum( vblock, "STACNT", &one, &ereal, (void *) stacnt )
       <= 0)
      || (c_getnum( vblock, "NITER", &one, &elong, (void *) niter )
       <= 0) )
	c_lasmsg( TRUE, "surface-getarg",
	  "Error getting program arguments" );
    if (c_getstr( vblock, "REGRESS", &one, &cmlen, argbuf ) <= 0)
	c_lasmsg( TRUE, "surface-getarg",
	  "Error getting program arguments" );
    if (strcmp( argbuf, "NONE" ) == 0)
	{
        *regress = 0;
	*slopeflag = TRUE;
	*slope = 0.0;
	}
    else if (strcmp( argbuf, "LINEAR" ) == 0)
        *regress = 1;
    else if (strcmp( argbuf, "NORMDIFF" ) == 0)
        *regress = 2;
    if ( (c_getnum( vblock, "MAXND", &one, &ereal, (void *) maxnd )
       <= 0)
      || (c_getnum( vblock, "MINFRACT", &one, &ereal,
       (void *) minfract ) <= 0)
      || (c_getnum( vblock, "THRESH", &one, &ereal, (void *) thresh )
       <= 0)
      || (c_getnum( vblock, "FILL", &one, &ereal, (void *) fill )
       <= 0) )
	c_lasmsg( TRUE, "surface-getarg",
	  "Error getting program arguments" );

/*  Retrieve TAE global variables  */

    if ( (c_getnum( vblock, "$MSGTIM", &one, &elong, (void *) pmsgtim )
       <= 0)
      || (c_getstr( vblock, "$DDRVFLG", &one, &cmlen, argbuf ) <= 0) )
	c_lasmsg( TRUE, "surface-getarg",
	  "Error getting program arguments" );
    if (strncmp (argbuf, "IN", 2) == 0)
	*ddrvflg = INVAL;
    else
	*ddrvflg = UNKNOW;

/*  Set up temporary file for optional header listing  */

    strcpy (prtfile,"surface");
    if (c_setprt (vblock, dummy, &ifpfile, dummy, prtfile, listfile)
      != E_SUCC) 
	c_lasmsg( TRUE, "surface-getarg",
	  "Error getting program arguments" );
    if (ifpfile == 0)
	listfile[0] = '\0';

    return;
    }			/* End of function srf_getarg.c */
