/*  srf_getsta.c -- Get station data and convert location coords

NAME:
    srf_getsta()

PURPOSE:
    Called by program SURFACE to read the file containing station
    locations, elevations,  and observation values and convert the
    locations to image coordinates.

ARGUMENTS:

    char    *inelev	(in)	Hostname of input elevation image.
    long    *window	(in)	Input elevation image window.
    char    *inobs	(in)	Name of station location and data file.
    float   *slope	(mod)	Slope of regression line between
				variable and elevation.
    long    *slopeflag	(mod)	Flag set true if slope specified
				explicitly or in observations file.
    FILE    *prtfp	(in)	File pointer for temporary report file.
    long    *obscnt	(out)	Number of observation stations.
    double   **obsptr	(out)	Pointer to array to contain, for each
				observation, line and sample relative to
				elevation image window, station
				elevation, and observation value.

RETURN VALUE:
    None.  Function aborts on fatal error.

FILE REFERENCES:
    <inobs>	READ to obtain location, elevation, and observation
		value for each station; may also contain constant slope
		of regression between observed variable and elevation.
    <prtfp>	WRITE to report values of slope and intercept.

PROGRAM HISTORY:                                  
    2001-06	Initial version, R. A. White, PSU/ESSC.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in surface.pdf.

ALGORITHM DESCRIPTION:
    Open station data file
    Enter name of station observation file into report file
    Write constant slope value to report file if given as arguments
    If first record is slope
	If value not entered as program argument
	    Use this value and add note to report file
	Else
	    Add warning line to report file
    Determine number of observation records and allocate data buffer
    Read in observation records
    Close station data file
    Get elevation grid ddr
    Initialize coordinate transformation package
    For each observation record
	Compute projection coordinates of station
	Convert projection coordinates to elevation image coordinates
    									
***************   End of Prolog for srf_getsta   ***************	*/
									/*
Header Files:			       					*/

#include	<ctype.h>	/* Character classification functions	*/
#include	<stdio.h>	/* Standard I/O functions		*/
#include	<string.h>	/* String manipulation functions	*/
#include	<time.h>	/* Time/date manipulation functions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"geompak.h"	/* LAS geometric transformation fcn's	*/
									/*
***************   Start of Function srf_getsta   ***************	*/

void srf_getsta(
    char    *inelev,	/* Hostname of input elevation image */
    long    *window,	/* Input elevation image window */
    char    *inobs,	/* Name of station location and data file */
    float   *slope,	/* Slope of regression line between variable and
			   elevation */
    long    *slopeflag,	/* Flag set true if slope specified explicitly
			   or in observations file */
    FILE    *prtfp,	/* File pointer for temporary report file */
    long    *obscnt,	/* Number of observation stations */
    double   **obsptr)	/* Pointer to array to contain, for each
			   observation, line and sample relative to
			   elevation image window, station elevation,
			   and observation value */
    {
	                                              			/*
Local Variable Definitions:						*/

    double	coords[2];	/* Projection coordinates */
    char	*bufptr;	/* Pointer into input buffer */
    struct DDR	ddr;		/* Elevation-image DDR */
    double	fields[4];	/* Fields from obs record */
    double	*obs;		/* Array of y, x, elevation, and
				   obs. value for each station */
    long	i;		/* Loop counter/index */
    long	indatum = 8;	/* Input projection spheroid:  GRS80 */
    double	inparm[15];	/* Input projection parameters */
    long	inproj = 0;	/* Input projection:  always GEO */
    long	inunit = 4;	/* Input proj units:  decimal degr. */
    time_t	now;		/* Current system time */
    char	obsbuf[80];	/* Input buffer for obs data */
    FILE	*obsfp;		/* File pointer for observations file */
    long	outunit;	/* Output projection units code */
    long	prtprm[2]	/* Print parms for c_transinit() */
		  = {FALSE, FALSE};
    char	*unitptr;	/* Pointer into unitstring */
    char	unitstring[28];	/* Strings for DDR unit names */
    long	zero = 0;	/* Constant value zero */

/*  Open station data file  */

    if ((obsfp = fopen( inobs, "r" )) == NULL)
	c_lasmsg( TRUE, "surface-obsopen",
	  "Error opening observations file for input" );

/*  Enter name of station into report file  */

    now = time( NULL );
    fprintf( prtfp, "SURFACE, %s", ctime( &now ) );
    fprintf( prtfp, "\nObservation data file is %s\n", inobs );

/*  Check for slope in first record  */    

    obsbuf[0] = '\0';
    while (strlen( obsbuf ) < 2)
	if (fgets( obsbuf, sizeof( obsbuf ) + 1, obsfp ) == NULL)
	    c_lasmsg( TRUE, "surface-obsread",
	      "Error reading observations file" );
    if (*slopeflag)
	fprintf( prtfp, "Using constant slope value = %.4f from %s\n",
 	 *slope, "argument SLOPE" );
    if ((bufptr = strstr( obsbuf, "slope" )) != NULL)
	{
	if ((bufptr = strstr( bufptr + 5, "=" )) != NULL)
	    bufptr++;
	if ((bufptr == NULL) || (sscanf( bufptr, "%lf", fields ) != 1))
	    c_lasmsg( TRUE, "surface-slope",
	      "Bad slope specification in observations file" );
	*slope = fields[0];
	if (! *slopeflag)
	    {
	    fprintf( prtfp, "Using constant slope value = %.4f %s\n",
	      *slope, "from observations file" );
	    *slopeflag = 1;
	    }
	else
	    fprintf( prtfp, "  -- ignoring value (%.4f) %s\n", *slope,
	      "from observations file" );
	}

/*  Determine number of observation records and allocate buffers  */

    rewind( obsfp );
    *obscnt = 0;
    while (fgets( obsbuf, sizeof( obsbuf ) + 1, obsfp ) != NULL)
	{
	if (sscanf( obsbuf, "%lf %lf %lf %lf", fields, fields + 1,
	  fields +  2, fields + 3 ) == 4)
	    (*obscnt)++;
	}
    if (! feof( obsfp ))
	c_lasmsg( TRUE, "surface-obsread",
	  "Error reading observations file" );
    if (*obscnt <= 0)
	c_lasmsg( TRUE, "surface-badobs", "Cannot interpret obs file" );
    rewind( obsfp );
    if ((obs = (double*) malloc( 4 * *obscnt * sizeof( double ) ))
       == NULL)
    	c_lasmsg( TRUE, "surface-malloc", "Memory allocation error");
    *obsptr = obs;

/*  Extract observation values and station locations  */
    
    for (i = 0; fgets( obsbuf, sizeof( obsbuf ) + 1, obsfp ) != NULL; )
	{
	if (sscanf( obsbuf, "%lf %lf %lf %lf", obs + i, obs + i + 1,
	  obs + i + 2, obs + i + 3 ) == 4)
	    i += 4;
	}
    if (! feof( obsfp ))
	c_lasmsg( TRUE, "surface-obsread",
	  "Error reading observations file" );
    fclose( obsfp );

/*  Get elevation grid ddr  */

    if (c_getddr( inelev, &ddr ) != E_SUCC)
	c_lasmsg( TRUE, "surface-fatal", "Fatal error encountered" );

/*  Initialize coordinate transformation package  */

    for (i = 0; i < 15; i++)
	inparm[i] = 0;
    strcpy( unitstring, "rad fee met sec deg dms    " );
    strncpy( unitstring + 24, ddr.proj_units, 3);
    unitptr = strstr( unitstring, unitstring + 24 );
    if (unitptr == NULL)
	c_lasmsg( TRUE, "surface-elevunit",
	  "Invalid projection units for elevation image:  %s",
	  ddr.proj_units );
    else
        outunit = (unitptr - unitstring) / 4;
    if (c_transinit( &inproj, &inunit, &zero, &indatum, inparm,
      &ddr.proj_code, &outunit, &ddr.zone_code, &ddr.datum_code,
      ddr.proj_coef, prtprm, (char *) NULL ) != E_SUCC)
	c_lasmsg( TRUE, "surface-fatal", "Fatal error encountered" );

/*  Convert station coordinate to elevation window coordinates  */

    for (i = 0; i < 4 * *obscnt; i += 4)
	{
	if (c_trans( &inproj, &inunit, &ddr.proj_code, &outunit,
	  obs + i + 1, obs + i, coords + 1 , coords) != E_SUCC)
	    c_lasmsg( TRUE, "surface-fatal",
	      "Fatal error encountered" ); 
	obs[i] = (ddr.upleft[0] - coords[0]) / ddr.pdist_y
	  - (window[0] - 1);
	obs[i + 1] = (coords[1] - ddr.upleft[1]) / ddr.pdist_x
	  - (window[1] - 1);
	}

    return;
    }			/* End of Function srf_getsta */
