/*  srf_interp.c -- Interpolate using simple linear elevation dependence

NAME:
    srf_interp()

PURPOSE:
    Called by SURFACE to apply a Gaussian weighting function with a
    cutoff distance to observations from nearby stations to estimate the
    value of a variable at each point of the output image, assuming that
    the variable has a linear dependence on elevation.  The cutoff
    distance is adjusted based on the local density of observation
    stations so that a roughly contant number of stations are used at
    each output point.  The cutoff distance is interpolated bilinearly
    from a coarse grid of cutoff distances.

ARGUMENTS:
    struct PARBLK
	    *vblock     (in)	Structure to receive arguments from TAE.
    char    *inelev	(in)	Hostname of input elevation image.
    long    window[4]	(in)	Input elevation image window.
    long    *band	(in)	Input elevation image band array.
    char    *out	(in)	Hostname of output image.
    long    odcode	(in)	Data type code for output image.
    float   alpha	(in)	Shape parameter of Gaussian weighting
				function.
    long    slopeflag	(in)	Flag set true if constant slope
				specified.
    float   slopeconst	(in)	Constant slope of regression line
				between variable and elevation.
    long    regress	(in)	Code for regression type.
    float   maxnd	(in)	Maximum absolute value of normalized
				difference.
    float   minfract	(in)	Minimum fraction of stations above
				threshold for non-fill output.
    float   thresh	(in)	Threshold for using variable value from
				station.
    float   fill	(in)	Fill value for output pixels failing
				MINFRACT test.
    double  *obs	(in)	Array containing, for each observation,
				line and sample relative to elevation
				image window, station elevation, and
				observation value.
    long    obscnt	(in)	Number of observation stations.
    float   *cutoff	(in)	Array of cutoff distances at each point
				on coarse grid.
    long    griddist	(in)	Grid spacing in coarse grid of cutoff
				distance.
    long    ncol	(in)	Number of cols in coarse grid of
				cutoffs.
    FILE    *prtfp	(in)	File pointer for temporary report file.
    long    pmsgtim	(in)	Time between processing progress
				messages.

RETURN VALUE:
    None.  Function aborts on error.

FILE REFERENCES:
    <inelev>.img	READ to get elevation at each output pixel.
    <out>.img, .ddr	CREATEd to receive output image.
    <*prtfp>		APPENDed to add to processing report.

PROGRAM HISTORY:                                  
    2001-07	Initial version, R. A. White, PSU/ESSC.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in surface.pdf.

EXTERNAL FUNCTIONS CALLED:

    srf_slope	Compute elevation regression slope.

ALGORITHM DESCRIPTION:
    Allocate space for array of station weights
    Allocate space for cutoff values in current row of coarse grid
    Open elevation and output images
    Allocate I/O buffers and initialize processing group
    Initialize weight extrema
    Initialize output pixel extrema
    Start processing progress messages
    For each line of elevation grid
	For each pixel of line
	    Clear output value and weight accumulators
	    Determine cutoff distance by interpolation from coarse grid
	    For each station
		If station within cutoff distance
		    Compute station weight
		    Add weight to weight accumulator
		    If not rejected by threshold test
			Add station to list of those to use
			Add weight to non-rejected weight accumulator
	    If weighted fraction of stations accepted below minimum
	        Use fill value for pixel
	    Else
	        Compute slope of elevation regression <srf_slope>
		For each station within cutoff distance
		    Adjust value of observed value for elevation
		    Add product of weight times value to output value
		Divide output value by weight accumulator
	    Check weight accumulator for new extrema
	Check output line for extrema
	Write output line
    Stop processing progress messages
    Close elevation and output images
    Write minimum and maximum weight total to report file
    Write minimum and maximum station count to report file
    Write minimum, maximum, and mean slope to report file
    									
***************   End of Prolog for SRF_INTERP   ***************	*/
									/*
Header Files:			       					*/

#include	<math.h>	/* Standard math functions		*/
#include	<stdio.h>	/* Standard I/O functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"pixman.h"	/* LAS pixel manipulation functions	*/
#include	"util.h"	/* LAS utility function definitions	*/
									/*
External Function Prototypes Not Included in Header Files:		*/

    float srf_slope( long regress, float maxnd, float thresh,
		     long stacnt, float *weights, double *obs,
		     long *obsndx );
									/*
External Variable Declarations:						*/

    extern long curent;		/* Entity for progress messages */
    extern long curcat;		/* Category for progress messages */

									/*
***************   Start of Function srf_interp()   ***************	*/

void srf_interp(
    struct PARBLK
	    *vblock,	/* Structure to receive arguments from TAE */
    char    *inelev,	/* Hostname of input elevation image */
    long    window[4],  /* Input elevation image window */
    long    *band,	/* Input elevation image band array */
    char    *out,	/* Hostname of output image */
    long    odcode,	/* Data type code for output image */
    float   alpha,	/* Shape parameter of Gaussian weighting
			   function */
    long    slopeflag,	/* Flag set true if constant slope specified */
    float   slopeconst,	/* Constant slope of regression line between
			   variable and elevation */
    long    regress,	/* Code for regression type */
    float   maxnd,	/* Maximum absolute value of normalized
			   difference */
    float   minfract,	/* Minimum fraction of stations above threshold
			   for non-fill output */
    float   thresh,	/* Threshold for using variable value from
			   station */
    float   fill,	/* Fill value for output pixels failing MINFRACT
			   test */
    double  *obs,	/* Array containing, for each observation, line
			   and sample relative to elevation image
			   window, station elevation, and observation
			   value */
    long    obscnt,	/* Number of observation stations */
    float   *cutoff,	/* Array of cutoff distances at each point on
			   coarse grid */
    long    griddist,	/* Grid spacing in coarse grid of cutoff dist */
    long    ncol,	/* Number of cols in coarse grid of cutoffs */ 
    FILE    *prtfp,	/* File pointer for temporary report file */
    long    pmsgtim)	/* Time between processing progress messages */

    {
	                                              			/*
Local Variable Definitions:						*/

    long	access;		/* File access code */
    float	*buffer;	/* I/O buffer */
    long	buffersz;	/* Size of I/O buffer, in bytes */
    long	col;		/* Column in coarse grid of cutoffs */
    long	cutndx;		/* Index into coarse grid of cutoffs */
    float	*cuts;		/* Array of cutoffs for current line */
    float	cutsqr;		/* Square of current cutoff distande */
    float	cval = 0.0;	/* Cutoff distance for current pixel */
    float	*dcdx;		/* Change in cutoff distance with x */
    float	*dcdy;		/* Change in cutoff distance with y */
    long	dcode;		/* Data type code for I/O buffers */
    float	*ddcdxdy;	/* Change in cutoff with x and y */
    float	eminalpha;	/* Value of exp( - alpha ) */
    float	distsqr;	/* Square of distance to obs station */
    float	elev;		/* Elevation of current pixel */
    struct FDESC
		*fdarray[3];	/* Image file descriptors */
    float	gausscoeff;	/* Coeff of dist squared in gaussian */ 
    struct GDESC
		*gdesc;		/* I/O group descriptor */
    long	i;		/* Loop index/counter */
    long	line;		/* Current line of output image */
    double	minmax[2];	/* Minimum and maximum output values */
    long	msgtype;	/* Progres message format code */
    float	nonrejw;	/* Total weight for non-rejected obs */
    float	normd;		/* Normalized difference between obs
				   value at output pixel and value at
				   station */ 
    long	*obsndx = NULL;	/* Indices into observation array */
    long	offset[2];	/* Input and output buffer offsets */
    long	one=1;		/* Constant value one */
    long	opt[2]={0, 0};	/* Option flags for image open */
    float	outval;		/* Value of current output pixel */
    long	pixel;		/* Current pixel of output line */
    float	slope = 0.0;	/* Local slope of value vs elevation */
    double	slopeacc = 0.0;	/* Sum of all slope values */
    long	slopecnt = 0;	/* Number of slope values computed */
    float	slopemax = 0.0;	/* Maximum slope value */
    float	slopemin = 0.0;	/* Minimum slope value */
    long	stacnt;		/* Stations for current pixel */ 
    long	stamax;		/* Maximum stations for any pixel */
    long	stamin;		/* Minimum stations for any pixel */
    double	unity=1.0;	/* Constant value 1.0 */
    float	*weights;	/* Weights for observing stations */
    float	weightacc;	/* Accumulator for weights */
    float	wmax;		/* Maxium weights for any point */
    float	wmin;		/* Minium weights for any point */
    float	x;		/* Floating equivalent of pixel coord */ 
    float	y;		/* Floating equivalent of line number */
    long	zero=0;		/* Constant value zero */

/*  Allocate space for station weights and indices for each pixel  */

    if ( ((weights = (float *) malloc( obscnt * sizeof( float ) ))
          == NULL)
      || ((obsndx = (long *) malloc( obscnt * sizeof( long ) ))
	  == NULL) )
	c_lasmsg( TRUE, "surface-malloc", "Memory allocation error");

/*  Allocate space for cutoff values in current row of coarse grid  */

    if ((cuts = (float *) malloc( 4 * ncol * sizeof( float ) )) == NULL)
	c_lasmsg( TRUE, "surface-malloc", "Memory allocation error");
    dcdx = cuts + ncol;
    dcdy = dcdx + ncol;
    ddcdxdy = dcdy + ncol;

/*  Open elevation and output images  */

    band[1] = 0;
    access = IREAD;
    dcode = EREAL;
    if (c_eopens( fdarray, inelev, band, window, window + 1, window + 2,
      window + 3, offset, &access, &dcode, &zero, opt ) != E_SUCC)
	c_lasmsg( TRUE, "surface-fatal", "Fatal error encountered" );
    access = IWRITE;
    opt[0] = odcode;
    if (c_eopens( fdarray + 1, out, &one, &one, &one, window + 2,
      window + 3, offset + 1, &access, &dcode, &zero, opt ) != E_SUCC)
	c_lasmsg( TRUE, "surface-fatal", "Fatal error encountered" );

/*  Allocate I/O buffers and initialize processing group  */

    fdarray[2] = 0;
    if ((buffer = (float *) c_ealloc( fdarray, &buffersz )) == NULL)
	c_lasmsg( TRUE, "surface-fatal", "Fatal error encountered" );
    if (c_egroup( fdarray, &gdesc, (unsigned char *) buffer, &buffersz )
      != E_SUCC)
	c_lasmsg( TRUE, "surface-fatal", "Fatal error encountered" );

/*  Initialize extrema and related variables  */

    wmax = stamax = 0;
    wmin = 10000.0;
    stamin = obscnt;
    if (slopeflag)
	slope = slopeconst;
    else
	{
	slopemax = -1.0e38;
	slopemin = - slopemax;
	slopeacc = 0.0;
	slopecnt = 0;
	}

    if (c_inmnmx( &dcode, &one, minmax, minmax + 1 ) != E_SUCC)
	c_lasmsg( TRUE, "surface-fatal", "Fatal error encountered" );

/*  Start processing progress messages  */

    curent = 0;
    if( pmsgtim > 0)
	{
	msgtype = GEN2;
	c_inittm( "SURFACE", "processed", "line", "", "", window + 2,

	  &zero, &pmsgtim, &msgtype );
	c_startm();
	}

/*  Process one line of data at a time  */

    eminalpha = exp( - alpha );
    if (c_estep( &gdesc ) != E_SUCC)
	c_lasmsg( TRUE, "surface-fatal", "Fatal error encountered" );
    for (line = 0; line < window[2]; line++, curent++)
	{

/*  Setup cutoff interpolation for new line */

	if (! (line % griddist))
	    {
	    cutndx = ncol *line / griddist;
	    for (col = 0; col < ncol - 1; col++, cutndx++)
		{
		cuts[col] = cutoff[cutndx];
		dcdx[col]
		  = (cutoff[cutndx + 1] - cutoff[cutndx]) / griddist; 
		dcdy[col]
		  = (cutoff[cutndx + ncol] - cutoff[cutndx]) / griddist;
		ddcdxdy[col] = ((cutoff[cutndx + ncol + 1]
		  - cutoff[cutndx + ncol]) / griddist -	dcdx[col])
		  / griddist;
		}
	    }
	else
	    {
	    for (col = 0; col < ncol - 1; col++)
		{
		cuts[col] += dcdy[col];
		dcdx[col] += ddcdxdy[col];
		}
	    }

/*  Initialize cutoff distance and exponent for next pixel on line  */

	y = line + 1;
	for (pixel = 0, col = -1; pixel < window[3]; pixel++)
	    {
	    cval = (pixel % griddist) ? cval + dcdx[col] : cuts[++col];
	    cutsqr = cval * cval;
	    if (cutsqr <= 0.01)
		c_lasmsg( TRUE, "surface-zerocut",
		  "Zero cutoff value for line %d, pixel %d", line + 1,
		  pixel + 1 ); 
	    gausscoeff = alpha / cutsqr;
	    weightacc = nonrejw = 0.0;
	    x = pixel + 1;
	    elev = buffer[offset[0] + pixel];
	    outval = 0.0;

/*  Determine which observations contribute  */

	    for (stacnt = 0, i = 0; i < 4 * obscnt; i += 4)
		{
		distsqr =  (x - obs[i + 1]) * (x - obs[i + 1])
		  + (y - obs[i]) * (y - obs[i]);
		if (distsqr <= cutsqr)
		    {
		    weights[stacnt]
		      = exp( - gausscoeff * distsqr ) - eminalpha;
		    weightacc += weights[stacnt];
		    if ((minfract <= 0.0) || (obs[i + 3] >= thresh))
			{
			obsndx[stacnt] = i;
			nonrejw += weights[stacnt++];
			}
		    }
		}

/*  Check whether enough stations have values above threshold  */

	    if ((weightacc <= 0.0) || (nonrejw / weightacc < minfract) )
		outval = fill;

/*  Compute slope of weighted value versus elevation regression  */

	    else
		{
		if (! (slopeflag || (regress == 0)) )
		    {
		    slope = srf_slope( regress, maxnd, thresh, stacnt,
		      weights, obs, obsndx );
		    slopeacc += slope;
		    slopecnt++;
		    }

/*  Get contributions of all observations to output value  */

		for (i = 0; i < stacnt; i++)
		    {
		    if (regress <= 1)
			outval += weights[i] * (obs[obsndx[i] + 3]
			  + slope * (elev - obs[obsndx[i] + 2]));
		    else if (regress == 2)
			{
			normd = slope * (elev - obs[obsndx[i] + 2]);
			if (normd > maxnd)
			    normd = maxnd;
			else if (normd < - maxnd)
			    normd = -maxnd;
			outval += weights[i]
			  * ((1 + normd) / (1 -normd))
			  * obs[obsndx[i]+ 3]; 
			}
		    }
		if (weightacc > 0.0)
		    outval /= weightacc;
		}

/*  Copy interpolated value to output buffer  */

	    if (odcode != EREAL)
		outval = floor( outval + 0.5 );
	    buffer[offset[1] + pixel] = outval;

/*  Check for extrema  */

	    if (weightacc < wmin)
		wmin = weightacc;
	    if (weightacc > wmax)
		wmax =  weightacc;
	    if ((! slopeflag) && (regress != 0))
		{
		if (slope < slopemin)
		    slopemin = slope;
		if (slope > slopemax)
		    slopemax = slope;
		}
	    if (stacnt < stamin)
		stamin = stacnt;
	    if (stacnt > stamin)
		stamax = stacnt;
	    }

/*  Check output line for extrema and then write it  */

	c_pxmnmx( (void *) (buffer + offset[1]), window + 3, minmax,
	  minmax + 1, &dcode );
	if (c_estep( &gdesc ) != E_SUCC)
	    c_lasmsg( TRUE, "surface-fatal",
	      "Fatal error encountered" );
	}

/*  Stop progress messages and close image files  */

    if( pmsgtim > 0)
	c_stoptm();
    c_egclse( &gdesc );
    c_upddr( vblock, (void *) inelev, out, &one, (void *) window,
      (void*) band, &one, &unity, &unity, &one, &zero );
    c_svmnmx( out, &one, minmax, minmax + 1, "YES");

/*  Report ranges of weights, stations used, and slopes  */

    fprintf( prtfp, "Interpolation to elevation surface\n" );
    fprintf( prtfp, "    Minimum and maximum weights:  %.2f, %.2f\n",
      wmin, wmax );
    fprintf( prtfp, "    Minimum and maximum stations:  %ld, %ld\n",
      stamin, stamax );
    if ((! slopeflag) && (regress != 0))
	{
	fprintf( prtfp, "    Minimum and maximum slope:  %.5g, %.5g\n",
	  slopemin, slopemax );
	if (slopecnt > 0)
	    fprintf( prtfp, "    Mean slope:  %.5g\n",
	      slopeacc / slopecnt );
	}
    else if (regress == 0)
	fprintf( prtfp, "    No elevation interpolation performed\n" );

    return;
    }			/* End of Function srf_interp() */
