/*  srf_slope.c -- Compute slope of elevation regression line

NAME:
    srf_slope()

PURPOSE:
    Called by function srf_interp() of program SURFACE to compute the
    slope of the regression line between the observed values reported by
    a set of stations and their elevations.

ARGUMENTS:
    long    regress	(in)	Code for regression type.
    float   maxnd	(in)	Maximum absolute value of normalized
				difference.
    float   thresh	(in)	Threshold for using variable value from
				station.
    long    stacnt	(in)	Number of observing stations for pixel.
    float   *weights	(in)	Weights for observing stations.
    double  *obs	(in)	Array containing, for each observation,
				line and sample relative to elevation
				image window, station elevation, and
				observation value.
    long    *obsndx	(in)	Index into observation array for each
				station to be used. 

RETURN VALUE:
    srf_slope() returns the slope of the regression line.

PROGRAM HISTORY:                                  
    2001-07	Initial version, R. A. White, PSU/ESSC.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.7 and 2.8.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in surface.pdf.

ALGORITHM DESCRIPTION:

    Initialize accumulators for obs times elev and elev squared
    For each station in list of stations to be used
	For each other station with lower index
	    If linear regression
		Increment sums for weighted regression
	    Else if normed difference regression
		If both observations are above threshold
		    Compute normed difference
		    Increment sums for weighted regression
    Compute slope

    NOTE:  Since actual regression requires summing over all pairs of
    stations (i, j), and must give same results if station numbering is
    reversed, terms which are odd multiples of (val[i] - val[j]) will
    be zero, so only quadratic terms are needed for regression.

***************   End of Prolog for srf_slope()   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O functions		*/
#include	<math.h>	/* Standard math functionsI		*/
#include 	"las.h"		/* Standard LAS include files		*/
									/*
***************   Start of Function srf_slope()   ***************	*/

float srf_slope(
    long    regress,	/* Code for regression type */
    float   maxnd,	/* Maximum absolute value of normalized
			   difference */
    float   thresh,	/* Threshold for using variable value from
			   station */
    long    stacnt,	/* Number of observing stations for pixel*/
    float   *weights,	/* Weights for observing stations */
    double  *obs,	/* Array containing, for each observation, line
			   and sample relative to elevation image
			   window, station elevation, and observation
			   value */
    long    *obsndx)	/* Index into observation array for each station
			   to be used */ 
    {
	                                              			/*
Local Variable Definitions:						*/

    float	elevdiff;	/* Difference between elevations */ 
    long	i;		/* Loop index/counter */
    long	j;		/* Loop index/counter */
    float	obsdiff;	/* Difference between obs. values */
    float	obssum;		/* Sum of observed values */
    float	slope;		/* Slope of regression line */
    float	xsqracc;	/* Sum of squares of x distances */
    float	xyacc;		/* Sum of products of x and y dist. */

/*  Accumulate sums needed to compute slope  */

    xyacc = xsqracc = 0.0;
    for (i = 1; i < stacnt; i++)
	for (j = 0; j < i; j++)
	    {
	    elevdiff = (obs[obsndx[i] + 2] - obs[obsndx[j] + 2]);
	    obsdiff = (obs[obsndx[i] + 3] - obs[obsndx[j] + 3]);

/*  Add to accumulators for linear regression  */

	    if (regress == 1)
		{
		xyacc += weights[i] * weights[j] * elevdiff * obsdiff;
		xsqracc += weights[i] * weights[j] * elevdiff *
		  elevdiff;
		}

/*  Add to accumulators for normalized difference  */

	    else if (regress == 2)
		{
		if ( (obs[obsndx[i] + 3] >= thresh)
		     && (obs[obsndx[j] + 3] >= thresh) )
		    {
		    obssum = obsdiff + 2 * obs[obsndx[j] + 3];
		    if (fabs( obssum ) < 1.0e-10 * fabs( obsdiff) )
			c_lasmsg( TRUE, "surface-normd", "%s %g and %g",
			  "Sum of observations is zero, values ",
			  obs[obsndx[i] + 3], obs[obsndx[j] + 3] );
		    else if (obssum != 0.0)
			obsdiff /= obssum;
		    if (obsdiff > maxnd)
			obsdiff = maxnd;
		    else if (obsdiff < - maxnd)
			obsdiff = - maxnd;
		    xyacc += weights[i] * weights[j] * elevdiff *
		      obsdiff;
		    xsqracc += weights[i] * weights[j] * elevdiff *
		      elevdiff;
		    }
		}
	    }

/*  Compute slope of regression line  */

    slope = (xsqracc > 0.0) ? xyacc / xsqracc : 0.0;

    return slope;
    }			/* End of Function srf_slope() */
