/*  surface.c -- Grid elevation-dependent observations

NAME:
    SURFACE

PURPOSE:
    Surface creates a uniformly spaced surface grid from values of an
    elevation-dependent variable recorded at a set of observing stations
    which are not uniformly distributed.  The contribution of stations
    near a given grid point are weighted by an exponential function of
    the distance from the point to the station, subject to a cutoff
    radius.  The value of the variable at each contributing observing
    station is adjusted to the elevation of the grid point using the
    slope of a regression line normally computed at each output pixel
    from a weighted least-squares fit to the elevation and observed
    value differences or normalized differences between every pair of
    stations within the cutoff distance; optionally, a constant slope
    may be specified as a program argument or in the input observations
    file.  Inclusion of an observing station for computation of
    regression lines and output pixels may be subject to a threshold
    test, in which case output pixels will be set to a fill value if the
    the weighted fraction of stations with values above the threshold is
    less than a specified minimum value.

ARGUMENTS:
    INELEV	Image containing elevation for each output grid point.
    INOBS	File containing location of each observing station and
		value of variable to be surfaced.  There must be one
		record for each station, containing four
		whitespace-separated fields giving the latitude and
		longitude, in decimal degrees, the elevation in meters,
		and the value of the observed variable as a real or
		integer value.  South latitudes and west longitudes must
		be indicated by negative values.  A constant value for
		the slope of the regression line between observed values 
		and elevation may be specified in the first record of
		this file, in the format "slope=XXX.XXXX", where the
		number of places on each side of the decimal point is
		arbitrary; this value is ignored if argument SLOPE is
		specified.
    OUT		Name of output image.  Will contain gridded, elevation
		adjusted values of observed varialbe.
    ODTYPE	Data type of output image.
    ALPHA	Shape parameter of Gaussian weighting function.
    STACNT	Target average number of stations for each grid point.
    NITER	Number of iterations for refining maximum distance.
    SLOPE	Constant value for slope of regression line between
		variable and elevation.  This overrides a slope value
		specified in the INOBS file.  If SLOPE is defaulted and
		no value is specified in the INOBS file, a weighted
		regression is computed for each output pixel.
    REGRESS	Type of regression of variable against elevation, none,
		linear, or normalized difference.
    MAXND	Maximum absolute value of normalized difference.
		Normalized differences exceeding MAXND will be replaced
		by MAXND, with the same sign.
    MINFRACT	Minimum fraction of stations above threshold for
		non-fill output.
    THRESH	Threshold for using variable value from station.
    FILL	Fill value for output pixels failing MINFRACT test.  
    PRINT	Destination(s) for processing report.
    PRTFILE	Name of print file, if any, for processing report.

FILE REFERENCES:
    <INELEV>.img, .ddr	Input elevation image
     .ddr
    <INOBS>		Input station and observation data file.
    <OUT>.img, .ddr,	Output image.
     .his
    <PRTFILE>		Output file containing processing report.

PROGRAM HISTORY:                                  
    2001-06	Initial version, R. A. White, PSU/ESSC.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in sufrace.pdf.

EXTERNAL FUNCTIONS CALLED:

(Program Specific Functions)
    srf_cutoff  Determine cutoff distance for weighting function.
    srf_getarg	Get program arguments from TAE interface.
    srf_getsta  Get station data and convert location coordinates.
    srf_interp	Create output image from station data.
    srf_slope	Compute elevation regression slope (via srf_interp).

    Calling Hierarchy for Program Specific Functions:

	surface
	    srf_getarg
	    srf_getsta
	    srf_cutoff
	    srf_interp
		srf_slope

ALGORITHM DESCRIPTION:
    Initialize TAE argument-passing interface.
    Get program arguments <srf_getarg>
    Open temproray report file
    Get station data and convert locations to image coords <srf_getsta>
    Determine cutoff distance for weighting function <srf_cutoff>
    Create interpoloated output image from station data <srf_interp>
    Close temporary report file
    Write history label
    Write completion message

ALGORITHM REFERENCES:
    The surfacing algorithm is taken from P. E. Thornton, S. W. Running,
    and M. A. White, "Generating surfaces of daily meteorological
    variables over large regions of complex terrain", Journal of
    Hydrology 190 (1997), pp. 214-251.
    									
***************   End of Prolog for SURFACE   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"util.h"	/* LAS utility function definitions	*/
									/*
External Function Prototypes Not Included in Header Files:		*/

    void srf_cutoff( long *window, double *obs, long obscnt, long niter,
		     long stacnt, float alpha, FILE *prtfp,
		     float **cutptr, long *griddist, long *ncol, long
		     *nrow);
    void srf_getarg( struct PARBLK *vblock, char *inelev, long *window,
		     long *band, char *inobs, char *out, long *odcode,
		     float *alpha, float *stacnt, long *niter,
		     float *slope, long *slopeflag, long *regress,
		     float *maxnd, float *minfract, float *thresh,
		     float *fill, char *prtfile, char *listfile,
		     long *pmsgtim, long *ddrvflg ); 
    void srf_getsta( char *inelev, long *window, char *inobs,
		     float *slope, long *slopeflag, FILE *prtfp,
		     long *obscnt, double **obsptr );
    void srf_interp( struct PARBLK *vblock, char *inelev,
		     long window[4], long *band, char *out, long odcode, 
		     float alpha, long slopeflag, float slope,
		     long regress, float maxnd, float minfract,
		     float thresh, float fill, double *obs,
		     long obscnt, float *cutoff, long griddist,
		     long ncol, FILE *prtfp, long pmsgtim ); 
									/*
***************   Main Routine for SURFACE   ***************		*/

int main()
    {
	                                              			/*
Local Variable Definitions:						*/

    float 	alpha;		/* Shape parameter of Gaussian weighting
				   function */
    long  	band[MAXBND];	/* Input elevation image band (array
				   needed for terminating 0, possible
				   erroneous entry of multiple bands) */
    float	*cutoff;	/* Array of cutoff distances at each
				   point on coarse grid */
    long  	ddrvflg;	/* DDR validity option */
    float	fill;		/* Fill value for output pixels failing
				   MINFRACT test */
    long	griddist;	/* Grid spacing in sparse grid of cutoff
				   distances */
    char  	inelev[CMLEN];	/* Hostname of input elevation image */
    char  	inobs[CMLEN];	/* station location and data file */
    char  	listfile[CMLEN]; /* Name of output listing file */
    float	maxnd;		/* Maximum absolute value of normalized
				   difference */
    float	minfract;	/* Minimum fraction of stations above
				   threshold for non-fill output */
    long	ncol;		/* Number of columns in sparse grid of
				   cutoff distances */ 
    long  	niter;		/* Number of iterations for refining
				   maximum distance */
    long	nrow;		/* Number of rows in sparse grid of
				   cutoff distances */
    double	*obs;		/* Data for each observing station, in
				   order line and sample relative to
				   elevation image window, station
				   elevation, and observation value */ 
    long	obscnt;		/* Total number of observations */
    long  	odcode;		/* Data type code for output image */
    long	one=1;		/* Constant one */
    char  	out[CMLEN];	/* Hostname of output image */
    long  	pmsgtim;	/* Time between progress messages */
    char  	prtfile[CMLEN]; /* Name of temporary print file */
    FILE	*prtfp;		/* File pointer for print file */
    long	regress;	/* Code for regression type */
    float 	slope;		/* Constant slope of regression line
				   between variable and elevation */
    long  	slopeflag;	/* Set if constant slope specified as
				   argument or in observations file */
    float 	stacnt;		/* Target average number of stations for
				   each grid point */
    long	stdlun;		/* Unit number for standard output */
    float	thresh;		/* Threshold for using variable value
				   from station */
    struct PARBLK
	  	vblock;		/* TAE argument block */
    long	window[4];  	/* Input elevation image window */
    long	zero=0;		/* Constant zero */

/*  Initialize TAE  */

    c_inlas (&vblock, &stdlun);

/*  Get program arguments  */

    srf_getarg( &vblock, inelev, window, band, inobs, out, &odcode,
      &alpha, &stacnt, &niter, &slope, &slopeflag, &regress, &maxnd,
      &minfract, &thresh, &fill, prtfile, listfile, &pmsgtim, &ddrvflg );

/*  Open temporary report file  */

    if ((prtfp = fopen( prtfile, "w" )) == NULL)
	c_lasmsg( TRUE, "surface-prtopen",
	  "Error opening temporary report file for output" );

/*  Get station observations and convert locations to image coords  */

    srf_getsta( inelev, window, inobs, &slope, &slopeflag, prtfp,
      &obscnt, &obs );

/*  Determine cutoff distance for weighting function  */

    srf_cutoff( window, obs, obscnt, niter, stacnt, alpha, prtfp,
      &cutoff, &griddist, &ncol, &nrow );

/*  Create interpolated output image from station data  */

    srf_interp( &vblock, inelev, window, band, out, odcode, alpha,
      slopeflag, slope, regress, maxnd, minfract, thresh, fill, obs,
      obscnt, cutoff, griddist, ncol, prtfp, pmsgtim);

/*  Close report file and write history label  */

    fclose( prtfp );
    c_hstout (&vblock, out, (void *) " ", &zero, "SURFACE");

/*  Write standard completion message  */

    c_complt ("SURFACE", (void *) out, &one);
    
    return 0;	/* To suppress compiler warning */

   }			/* End of Program surface.c */
